.MathJax .mi, .MathJax .mo {
    color: inherit;
}

pre, kdb, samp, code, .rst-content tt, .rst-content code {
    word-break: inherit;
}

code, .rst-content tt, .rst-content code {
    white-space: pre-wrap;
}

#right-column .header div {
    float: left;
}
